/*
See LICENSE folder for this sample’s licensing information.

Abstract:
Main view controller for the AR experience.
*/

import RealityKit
import ARKit
import Vision
import SceneKit

extension CVPixelBuffer {
    func value(column: Int, row: Int) -> Float? {
        guard CVPixelBufferGetPixelFormatType(self) == kCVPixelFormatType_DepthFloat32 else { return nil }
        CVPixelBufferLockBaseAddress(self, .readOnly)
        if let baseAddress = CVPixelBufferGetBaseAddress(self) {
            let width = CVPixelBufferGetWidth(self)
            let index = column + row*width
            let offset = index * MemoryLayout<Float>.stride
            let value = baseAddress.load(fromByteOffset: offset, as: Float.self)
                CVPixelBufferUnlockBaseAddress(self, .readOnly)
            return value
        }
        CVPixelBufferUnlockBaseAddress(self, .readOnly)
        return nil
    }
}



class ViewController: UIViewController, ARSessionDelegate {
    
    @IBOutlet var arView: ARView!
    @IBOutlet weak var hideMeshButton: UIButton!
    @IBOutlet weak var resetButton: UIButton!
    @IBOutlet weak var planeDetectionButton: UIButton!
    @IBOutlet weak var debugInfosLabel: UILabel!
    
    let coachingOverlay = ARCoachingOverlayView()
    
    // Cache for 3D text geometries representing the classification values.
    var modelsForClassification: [ARMeshClassification: ModelEntity] = [:]

    private var handPoseRequest = VNDetectHumanHandPoseRequest()
    
//    private var overlayLayer = CAShapeLayer()
//    private var pointsPath = UIBezierPath()
    
    var resultAnchor = AnchorEntity()
    
    
    /// - Tag: ViewDidLoad
    override func viewDidLoad() {
        super.viewDidLoad()
        
        arView.session.delegate = self
        
        setupCoachingOverlay()

        arView.environment.sceneUnderstanding.options = []
        arView.debugOptions.insert(.showSceneUnderstanding)
        
        // For performance, disable render options that are not required for this app.
        arView.renderOptions = [.disableDepthOfField, .disableMotionBlur]
        
        // Manually configure what kind of AR session to run since
        // ARView on its own does not turn on mesh classification.
        //arView.automaticallyConfigureSession = false
        let configuration = ARWorldTrackingConfiguration()
        
        configuration.planeDetection = .horizontal
        configuration.sceneReconstruction = .mesh
        configuration.frameSemantics.insert(.personSegmentationWithDepth)
        configuration.environmentTexturing = .automatic
        
        arView.session.run(configuration)
        
        resultAnchor = AnchorEntity(.plane(.horizontal, classification: .any, minimumBounds: SIMD2<Float>(0.0,0.0))) //(world: SIMD3<Float>(0.0,0.0,0.0))
        resultAnchor.addChild(sphere(radius: 0.02, color: .blue))
        arView.scene.addAnchor(resultAnchor)
        
        handPoseRequest.maximumHandCount = 1
        
//        let tapRecognizer = UITapGestureRecognizer(target: self, action: #selector(handleTap(_:)))
//        arView.addGestureRecognizer(tapRecognizer)
    }
    
    var indexTip: CGPoint = CGPoint(x: 0,y: 0)
    
    
    func session(_ session: ARSession, didUpdate frame: ARFrame) {
        
        let orientation: CGImagePropertyOrientation = CGImagePropertyOrientation(rawValue: UInt32(UIDevice.current.orientation.rawValue))!
        let handler = VNImageRequestHandler(cvPixelBuffer: frame.capturedImage, orientation: orientation)
        do {
            // Perform VNDetectHumanHandPoseRequest
            try handler.perform([handPoseRequest])
            // Continue only when a hand was detected in the frame.
            // Since we set the maximumHandCount property of the request to 1, there will be at most one observation.
            guard let observation = handPoseRequest.results?.first else {
                return
            }
            // Get points for thumb and index finger.
            let thumbPoints = try observation.recognizedPoints(.thumb)
            let indexFingerPoints = try observation.recognizedPoints(.indexFinger)
            // Look for tip points.
            guard let thumbTipPoint = thumbPoints[.thumbTip], let indexTipPoint = indexFingerPoints[.indexTip] else {
                return
            }
            // Ignore low confidence points.
            guard thumbTipPoint.confidence > 0.3 && indexTipPoint.confidence > 0.3 else {
                return
            }
            // Convert points from Vision coordinates to AVFoundation coordinates.
            // top right of the screen
            //thumbTip = CGPoint(x: thumbTipPoint.location.x, y: 1 - thumbTipPoint.location.y)
            
            // Convert to top right screen coordinates with Y to the left, X down
            indexTip = CGPoint(x:(indexTipPoint.location.x) * CGFloat(arView.bounds.width),
                               y:(1 - indexTipPoint.location.y) * CGFloat(arView.bounds.height))
            
            print("\nIndexTip   - x = \((indexTip.x*100).rounded() / 100)/\(arView.bounds.width) - y = \((indexTip.y*100).rounded() / 100)/\(arView.bounds.height)")
            
            // body or depthdata // frame.estimatedDepthData // frame.sceneDepth?.depthMap
            if let segmentationBuffer:CVPixelBuffer = frame.estimatedDepthData {

                let segmentationWidth = CVPixelBufferGetWidth(segmentationBuffer)
                let segmentationHeight = CVPixelBufferGetHeight(segmentationBuffer)
                
                // !!! bizarrement la AR View bounds.width a l'air de faire la meme taille que le buffer
                let xConverted:CGFloat = indexTip.x * CGFloat(segmentationWidth) / CGFloat(arView.bounds.width)
                let yConverted:CGFloat = indexTip.y * CGFloat(segmentationHeight) / CGFloat(arView.bounds.height)
                
//                print("segCols = \(segmentationCols) + segRows: \(segmentationRows)")
//                let colPosition = indexTip.x / arView.bounds.height * CGFloat(segmentationCols)
//                //UIScreen.main.bounds.width arView.bounds.width frame.camera.imageResolution.width
//                let rowPosition = indexTip.y / arView.bounds.width * CGFloat(segmentationRows)
//                print("colPo: \(colPosition) - rowPo: \(rowPosition)")
                
                print("Seg Buffer - x: \(Int(xConverted))/\(segmentationWidth) y: \(Int(yConverted))/\(segmentationHeight)")

                if let indexDepth:Float = segmentationBuffer.value(column: Int(xConverted), row: Int(yConverted)) {
                    
                    print("Finger Depth: \((indexDepth*100).rounded())cm")
                    
                    if indexDepth != 0 {
                        let cameraIntrinsics = frame.camera.intrinsics
                        
                        //let indexTipBisx:Float = Float(CGFloat(segmentationWidth) - indexTip.x)
                        //let indexTipBisy:Float = Float(CGFloat(segmentationHeight) - indexTip.y)

                        var xrw: Float = (Float(indexTip.x) - cameraIntrinsics[2][0]) * indexDepth
                        xrw = xrw / cameraIntrinsics[0][0]
                        var yrw: Float = (Float(indexTip.y) - cameraIntrinsics[2][1]) * indexDepth
                        yrw = yrw / cameraIntrinsics[1][1]
                        let xyzw: SIMD4<Float> = SIMD4<Float>(xrw, yrw, indexDepth, 1.0)
//                        print("xyzw: \(xyzw)")
//                        print("viewMatrix: \(frame.camera.viewMatrix(for: .portrait))")
                        let vecResult = frame.camera.viewMatrix(for: .portrait) * xyzw
                        print("vecResult: \(vecResult)")

                        debugInfosLabel.numberOfLines = 0
                        debugInfosLabel.text = "IndexTip   - x = \((indexTip.x*100).rounded() / 100)/\(arView.bounds.width) - y = \((indexTip.y*100).rounded() / 100)/\(arView.bounds.height)\nFinger Depth: \((indexDepth*100).rounded())cm\n"
                        
                        resultAnchor.setPosition(SIMD3<Float>(vecResult.x, vecResult.y, vecResult.z), relativeTo: nil)
                    
                    }
                }

            }
        }
        catch {
                print("Error: Vision request failed with error \"\(error)\"")
        }
        
    }
    
    
    
    func screenToWorld(screenPoint: CGPoint, depth: Float, screenSize: CGSize, projectionMatrix: simd_float4x4, viewMatrix:simd_float4x4) -> SIMD3<Float>{

        let viewProj = projectionMatrix * viewMatrix
        let inverse = simd_inverse(viewProj)

        let x = (2.0 * screenPoint.x ) / screenSize.width - 1.0
        let y = 1.0 - (2.0 * screenPoint.y) / screenSize.height

        let inPoint = SIMD4<Float>(Float(x), Float(y), depth * 2.0 - 1.0, 1.0)
        
        var position = inverse * inPoint
        position.w = 1.0 / position.w

        position.x *= position.w
        position.y *= position.w
        position.z *= position.w

        return SIMD3<Float>(position.x, position.y, position.z)
    }
    
    
    func processPoints(indexTip: CGPoint?) {
        if indexTip != nil {
            showPoints([indexTip!], color: .blue)
            return
        }
    }
    
    var path = UIBezierPath()
    
    
    
    func showPoints(_ points: [CGPoint], color: UIColor) {
        //pointsPath.removeAllPoints()
        for point in points {
//            pointsPath.move(to: point)
//            pointsPath.addArc(withCenter: point, radius: 15, startAngle: 0, endAngle: 2 * .pi, clockwise: true)
            
            path = UIBezierPath()
            path = UIBezierPath(ovalIn: CGRect(x: point.x, y: point.y, width: 10, height: 10))
            UIColor.yellow.setStroke()
            UIColor.red.setFill()
            path.lineWidth = 5
            path.stroke()
            path.fill()
        }
//        overlayLayer.fillColor = color.cgColor
//        CATransaction.begin()
//        CATransaction.setDisableActions(true)
//        overlayLayer.path = pointsPath.cgPath
//        CATransaction.commit()
    }
    
//    override func layoutSublayers(of layer: CALayer) {
//        super.layoutSublayers(of: layer)
//        if layer == previewLayer {
//            overlayLayer.frame = layer.bounds
//        }
//    }
//
//    private func setupOverlay() {
//        previewLayer.addSublayer(overlayLayer)
//    }
    
    
    
    
    
    
    
    
    
    
    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        // Prevent the screen from being dimmed to avoid interrupting the AR experience.
        UIApplication.shared.isIdleTimerDisabled = true
    }
    
    override var prefersHomeIndicatorAutoHidden: Bool {
        return true
    }
    
    override var prefersStatusBarHidden: Bool {
        return true
    }
    
    /// Places virtual-text of the classification at the touch-location's real-world intersection with a mesh.
    /// Note - because classification of the tapped-mesh is retrieved asynchronously, we visualize the intersection
    /// point immediately to give instant visual feedback of the tap.
    @objc
    func handleTap(_ sender: UITapGestureRecognizer) {
        // 1. Perform a ray cast against the mesh.
        // Note: Ray-cast option ".estimatedPlane" with alignment ".any" also takes the mesh into account.
        let tapLocation = sender.location(in: arView)
        if let result = arView.raycast(from: tapLocation, allowing: .estimatedPlane, alignment: .any).first {
            // ...
            // 2. Visualize the intersection point of the ray with the real-world surface.
            let resultAnchor = AnchorEntity(world: result.worldTransform)
            resultAnchor.addChild(sphere(radius: 0.01, color: .lightGray))
            arView.scene.addAnchor(resultAnchor, removeAfter: 3)

            // 3. Try to get a classification near the tap location.
            //    Classifications are available per face (in the geometric sense, not human faces).
            nearbyFaceWithClassification(to: result.worldTransform.position) { (centerOfFace, classification) in
                // ...
                DispatchQueue.main.async {
                    // 4. Compute a position for the text which is near the result location, but offset 10 cm
                    // towards the camera (along the ray) to minimize unintentional occlusions of the text by the mesh.
                    let rayDirection = normalize(result.worldTransform.position - self.arView.cameraTransform.translation)
                    let textPositionInWorldCoordinates = result.worldTransform.position - (rayDirection * 0.1)
                    
                    // 5. Create a 3D text to visualize the classification result.
                    let textEntity = self.model(for: classification)

                    // 6. Scale the text depending on the distance, such that it always appears with
                    //    the same size on screen.
                    let raycastDistance = distance(result.worldTransform.position, self.arView.cameraTransform.translation)
                    textEntity.scale = .one * raycastDistance

                    // 7. Place the text, facing the camera.
                    var resultWithCameraOrientation = self.arView.cameraTransform
                    resultWithCameraOrientation.translation = textPositionInWorldCoordinates
                    let textAnchor = AnchorEntity(world: resultWithCameraOrientation.matrix)
                    textAnchor.addChild(textEntity)
                    self.arView.scene.addAnchor(textAnchor, removeAfter: 3)

                    // 8. Visualize the center of the face (if any was found) for three seconds.
                    //    It is possible that this is nil, e.g. if there was no face close enough to the tap location.
                    if let centerOfFace = centerOfFace {
                        let faceAnchor = AnchorEntity(world: centerOfFace)
                        faceAnchor.addChild(self.sphere(radius: 0.01, color: classification.color))
                        self.arView.scene.addAnchor(faceAnchor, removeAfter: 3)
                    }
                }
            }
        }
    }
    
    @IBAction func resetButtonPressed(_ sender: Any) {
        if let configuration = arView.session.configuration {
            arView.session.run(configuration, options: .resetSceneReconstruction)
        }
    }
    
    @IBAction func toggleMeshButtonPressed(_ button: UIButton) {
        let isShowingMesh = arView.debugOptions.contains(.showSceneUnderstanding)
        if isShowingMesh {
            arView.debugOptions.remove(.showSceneUnderstanding)
            button.setTitle("Show Mesh", for: [])
        } else {
            arView.debugOptions.insert(.showSceneUnderstanding)
            button.setTitle("Hide Mesh", for: [])
        }
    }
    
    /// - Tag: TogglePlaneDetection
    @IBAction func togglePlaneDetectionButtonPressed(_ button: UIButton) {
        guard let configuration = arView.session.configuration as? ARWorldTrackingConfiguration else {
            return
        }
        if configuration.planeDetection == [] {
            configuration.planeDetection = [.horizontal, .vertical]
            button.setTitle("Stop Plane Detection", for: [])
        } else {
            configuration.planeDetection = []
            button.setTitle("Start Plane Detection", for: [])
        }
        arView.session.run(configuration)
    }
    
    func nearbyFaceWithClassification(to location: SIMD3<Float>, completionBlock: @escaping (SIMD3<Float>?, ARMeshClassification) -> Void) {
        guard let frame = arView.session.currentFrame else {
            completionBlock(nil, .none)
            return
        }
    
        var meshAnchors = frame.anchors.compactMap({ $0 as? ARMeshAnchor })
        
        // Sort the mesh anchors by distance to the given location and filter out
        // any anchors that are too far away (4 meters is a safe upper limit).
        let cutoffDistance: Float = 4.0
        meshAnchors.removeAll { distance($0.transform.position, location) > cutoffDistance }
        meshAnchors.sort { distance($0.transform.position, location) < distance($1.transform.position, location) }

        // Perform the search asynchronously in order not to stall rendering.
        DispatchQueue.global().async {
            for anchor in meshAnchors {
                for index in 0..<anchor.geometry.faces.count {
                    // Get the center of the face so that we can compare it to the given location.
                    let geometricCenterOfFace = anchor.geometry.centerOf(faceWithIndex: index)
                    
                    // Convert the face's center to world coordinates.
                    var centerLocalTransform = matrix_identity_float4x4
                    centerLocalTransform.columns.3 = SIMD4<Float>(geometricCenterOfFace.0, geometricCenterOfFace.1, geometricCenterOfFace.2, 1)
                    let centerWorldPosition = (anchor.transform * centerLocalTransform).position
                     
                    // We're interested in a classification that is sufficiently close to the given location––within 5 cm.
                    let distanceToFace = distance(centerWorldPosition, location)
                    if distanceToFace <= 0.05 {
                        // Get the semantic classification of the face and finish the search.
                        let classification: ARMeshClassification = anchor.geometry.classificationOf(faceWithIndex: index)
                        completionBlock(centerWorldPosition, classification)
                        return
                    }
                }
            }
            
            // Let the completion block know that no result was found.
            completionBlock(nil, .none)
        }
    }
    
    func session(_ session: ARSession, didFailWithError error: Error) {
        guard error is ARError else { return }
        let errorWithInfo = error as NSError
        let messages = [
            errorWithInfo.localizedDescription,
            errorWithInfo.localizedFailureReason,
            errorWithInfo.localizedRecoverySuggestion
        ]
        let errorMessage = messages.compactMap({ $0 }).joined(separator: "\n")
        DispatchQueue.main.async {
            // Present an alert informing about the error that has occurred.
            let alertController = UIAlertController(title: "The AR session failed.", message: errorMessage, preferredStyle: .alert)
            let restartAction = UIAlertAction(title: "Restart Session", style: .default) { _ in
                alertController.dismiss(animated: true, completion: nil)
                self.resetButtonPressed(self)
            }
            alertController.addAction(restartAction)
            self.present(alertController, animated: true, completion: nil)
        }
    }
        
    func model(for classification: ARMeshClassification) -> ModelEntity {
        // Return cached model if available
        if let model = modelsForClassification[classification] {
            model.transform = .identity
            return model.clone(recursive: true)
        }
        
        // Generate 3D text for the classification
        let lineHeight: CGFloat = 0.05
        let font = MeshResource.Font.systemFont(ofSize: lineHeight)
        let textMesh = MeshResource.generateText(classification.description, extrusionDepth: Float(lineHeight * 0.1), font: font)
        let textMaterial = SimpleMaterial(color: classification.color, isMetallic: true)
        let model = ModelEntity(mesh: textMesh, materials: [textMaterial])
        // Move text geometry to the left so that its local origin is in the center
        model.position.x -= model.visualBounds(relativeTo: nil).extents.x / 2
        // Add model to cache
        modelsForClassification[classification] = model
        return model
    }
    
    func sphere(radius: Float, color: UIColor) -> ModelEntity {
        let sphere = ModelEntity(mesh: .generateSphere(radius: radius), materials: [SimpleMaterial(color: color, isMetallic: false)])
        // Move sphere up by half its diameter so that it does not intersect with the mesh
        sphere.physicsBody = nil
        //sphere.position.y = radius
        return sphere
    }
}
